/*
 * Test program to demonstrate interaction between RexxDW and Javascript
 * via Regina's rxstack which can operate as a websockets server
 */
Trace O
Numeric Digits 15
--OPTIONS SINGLE_INTERPRETER
Call RxFuncAdd 'DW_LoadFuncs', 'rexxdw', 'DW_LoadFuncs'
Call dw_loadfuncs

!globalv = '!REXXDW. !global.'

Call dw_init

!global.!windowstyle = dw_or( !REXXDW.!DW_FCF_SYSMENU, !REXXDW.!DW_FCF_TITLEBAR, !REXXDW.!DW_FCF_SHELLPOSITION, !REXXDW.!DW_FCF_TASKLIST, !REXXDW.!DW_FCF_DLGBORDER )

!global.!allowcallbacks = 0

!global.!screen_width = dw_screen_width()
!global.!screen_height = dw_screen_height()
!global.!buttonheight = 25
!global.!crlf = d2c(13)||d2c(10)

stationlatitude ='-27.728033'
stationlongitude = '152.983486'
latitude ='-27.728033'
longitude = '152.983486'
address = '133 Teviot Road, Greenbank'

-- build a dialog showing the address and lat long and ask if this is to saved
win = dw_window_new( !REXXDW.!DW_DESKTOP, 'Check Location', dw_or( !global.!windowstyle, !REXXDW.!DW_FCF_SIZEBORDER, !REXXDW.!DW_FCF_MINMAX ) )
Call dw_signal_connect win, !REXXDW.!DW_DELETE_EVENT, 'QuitCallback', win
say 'main:' win
topbox = dw_box_new( !REXXDW.!DW_VERT )
Call dw_box_pack_start win, topbox, 0, 0, !REXXDW.!DW_EXPAND_HORZ, !REXXDW.!DW_EXPAND_VERT, 0
/*
 * Add a horz box for text entries
 */
data_box = dw_box_new( !REXXDW.!DW_VERT )
Call dw_box_pack_start topbox, data_box, 0, 0, !REXXDW.!DW_EXPAND_HORZ, !REXXDW.!DW_EXPAND_VERT, 0
-- Google Map
html = dw_html_new( 340 )
Call dw_box_pack_start data_box, html, 550, 650, !REXXDW.!DW_DONT_EXPAND_HORZ, !REXXDW.!DW_DONT_EXPAND_VERT, 0
-- create our queue on rxstack
saveq = rxqueue( 'Get' )
Call rxqueue 'Delete', 'BASS@localhost'
Call rxqueue 'Create', 'BASS@localhost'
Call rxqueue 'Set', 'BASS@localhost'
Queue stationlatitude'|'stationlongitude'|'latitude'|'longitude'|'location
-- build the map with a marker
Call dw_html_raw html, GetLocationChecker()
/*
 * Create Cancel and Apply buttons
 */
buttonbox = dw_box_new( !REXXDW.!DW_HORZ )
Call dw_box_pack_start topbox, buttonbox, 0, !global.!buttonheight+10, !REXXDW.!DW_EXPAND_HORZ, !REXXDW.!DW_DONT_EXPAND_VERT, 5

b1 = dw_button_new( 'Ok', 0 )
Call dw_box_pack_start buttonbox, b1, 0, !global.!buttonheight, !REXXDW.!DW_EXPAND_HORZ, !REXXDW.!DW_DONT_EXPAND_VERT, 5
Call dw_signal_connect b1, !REXXDW.!DW_CLICKED_EVENT, 'CheckLocationApplyCallback', win, saveq
/*
 * Display the window
 */
Call dw_window_set_gravity win, !REXXDW.!DW_GRAV_CENTRE, !REXXDW.!DW_GRAV_CENTRE
Call dw_window_set_pos win, 0, 0
Call dw_window_show win
Call dw_main_sleep 10

!global.!allowcallbacks = 1

Call dw_main

Return 0

CheckLocationApplyCallback:
Parse Arg ., win, saveq
Do Queued()
   Parse Pull line
End
If Strip( line ) \= '' Then
   Do
      Parse Var line lat '|' long '|' address
      Call dw_messagebox 'Result in Rexx/DW', dw_or( !REXXDW.!DW_MB_OK, !REXXDW.!DW_MB_WARNING ), 'Address:' address 'Lat:' lat 'Long:' long
   End
Return 0

QuitCallback:
If !global.!allowcallbacks = 0 Then Return 1
Parse Arg win, mainwin
Call dw_window_destroy mainwin
Call dw_exit 0
Return

GetLocationChecker: Procedure Expose (!globalv)
start = .LINE + 2
/*
<!DOCTYPE html>
<html>
<head>
<meta name="viewport" content="initial-scale=1.0, user-scalable=no" />
<script src="https://maps.google.com/maps/api/js?key=AIzaSyAlMex9agl1ane4sSzSW1EEVLbXKUDjnUw"></script>
<script>
var geocoder
var socket
var action
var map,marker
var newAddress, newLat, newLong

function geocodePosition( pos, queueData )
{
  geocoder.geocode(
  {
    latLng: pos
  }, function(responses)
  {
    if (responses && responses.length > 0)
    {
      updateMarkerAddress(responses[0].formatted_address)
      updateMarkerPosition( pos )
    }
    else
    {
      updateMarkerAddress("Cannot determine address at this location.")
    }
    if ( queueData == true )
       sendData()
  })
}

function geocodeAddress( address, queueData )
{
  geocoder.geocode(
  {
    address: address
  }, function(responses)
  {
    if (responses && responses.length > 0)
    {
      map.setCenter( responses[0].geometry.location )
      marker.setPosition( responses[0].geometry.location )
      updateMarkerPosition( responses[0].geometry.location )
      updateMarkerAddress( address )
    }
    else
    {
      updateMarkerAddress("Cannot determine position at this address.")
    }
    if ( queueData == true )
       sendData()
  })
}

function updateMarkerPosition(latLng)
{
  document.getElementById("info").innerHTML = [
    latLng.lat().toFixed(6),
    latLng.lng().toFixed(6)
  ].join(", ")
  newLat = latLng.lat().toFixed(6)
  newLong = latLng.lng().toFixed(6)
}

function updateMarkerAddress(str)
{
   document.getElementById("address").innerHTML = str
   newAddress = str
}

function sendData()
{
   var str = "/fifo:BASS:" + newLat + "|" + newLong + "|" + newAddress
   socket.send( str )
}

function initialize()
{
   // get address/lat/long from rxstack
   geocoder = new google.maps.Geocoder()
   socket = new WebSocket( 'ws://localhost:5757', 'rxstack')
   socket.onopen = function ()
   {
      socket.send( "/pull:BASS" )
   }
   socket.onmessage = function (event)
   {
      var data = event.data.slice( 2 )
      if ( data != "" )
      {
         var item = data.split( "|" )
         var stationLatLng, latLng
         // default to supplied station
         if ( item[0] != "" && item[1] != "" )
         {
            stationLatLng = new google.maps.LatLng( item[0], item[1] )
         }
         else
         {
            stationLatLng = new google.maps.LatLng( "-27.728033", "152.983486" )
         }
         // create map with center at station
         map = new google.maps.Map(document.getElementById("mapCanvas"),
         {
            zoom: 15,
            center: stationLatLng,
            mapTypeId: google.maps.MapTypeId.ROADMAP
         })
         marker = new google.maps.Marker(
         {
            position: stationLatLng,
            title: "Incident",
            map: map,
            draggable: true
         })
         // Update current position info.
         updateMarkerPosition( stationLatLng )
         geocodePosition( stationLatLng, false )

         if ( item[2] != "" && item[3] != "" )
         {
            // we have a lat and long for incident
            latLng = new google.maps.LatLng( item[2], item[3] )
            map.setCenter( latLng )
            marker.setPosition( latLng )
            geocodePosition( latLng, true )
         }
         else if ( item[4] != "" )
         {
            // we have an address only
            geocodeAddress( item[4], true )
         }
         google.maps.event.addListener(marker, "dragend", function()
         {
            updateMarkerPosition( marker.getPosition() )
            geocodePosition( marker.getPosition(), true )
         })
      }
   }
}

// Use onload handler to fire off the app.
</script>
</head>
<body onload="initialize()">
  <style>
  #mapCanvas {
    width: 500px;
    height: 500px;
  }
  #infoPanel {
    margin-top: 10px;
  }
  #infoPanel div {
    margin-bottom: 5px;
  }
  </style>

  <div>Drag marker to incident location</div>
  <div id="mapCanvas"></div>
  <div id="infoPanel">
    <b>Current position:</b>
    <div id="info"></div>
    <b>Closest matching address:</b>
    <div id="address"></div>
  </div>
</body>
</html>
*/
end = .LINE - 2
html = ""
Do i = start To end
   html = html||Sourceline(i)||!global.!crlf
End
Return html
