#!/usr/bin/regina
/*
 * This program controls a presentation
 */
Trace O
Signal on Halt Name HaltProgram
Signal On Novalue
Numeric Digits 30
!globalv = '!REXXWS. !REXXJSON. !global. !key.'

Call SetupKeyCodes
Call GetScreenDimensions

Call RxFuncAdd 'WSLoadFuncs', 'rexxws', 'WSLoadFuncs'
Call WSLoadFuncs
Call RxFuncAdd 'JSONLoadFuncs', 'rexxjson', 'JSONLoadFuncs'
Call JSONLoadFuncs

-- constants
!global.!debuglog = 0
!global.!port = '8100'
!global.!server_home = Directory()'/'
!global.!crlf = '0d0a'x
!global.!buffersize = 10000
-- global variables
!global.!slide = 0
!global.!slidewsi = 0

Parse Source os . me
Call Directory !global.!server_home
If os = 'WIN32' | os = 'WIN64' Then !global.!sep = '\'
Else !global.!sep = '/'

Call DoLog 'Starting talk-server.rexx: Rexx/WS version:' WSVariable( 'VERSION' ) 'libwebsockets version:' WSVariable( 'WSVERSION' ), 1
Call DoLog 'Listening on port' !global.!port 'in' !global.!server_home, 1
If !global.!debuglog = 1 Then logging = WSOr(!REXXWS.!LOG_DEBUG,!REXXWS.!LOG_INFO,!REXXWS.!LOG_NOTICE,!REXXWS.!LOG_ERR,!REXXWS.!LOG_WARN )
Else logging = 0
Call DoLog 'Screen dimension:' !global.!screenwidth 'x' !global.!screenheight, 1
Call WSSetLogLevel( logging )

rc = WSAddProtocol( !REXXWS.!PROTOCOL_HTTP, 'HTTPCallback' )
rc = WSAddProtocol( !REXXWS.!PROTOCOL_WS,'SlideCallback', 'slide', !global.!buffersize )
rc = WSAddProtocol( !REXXWS.!PROTOCOL_WS,'ControlCallback', 'control', !global.!buffersize )

options = 0
!global.!context = WSCreateContext( !global.!port, '', '', options ) -- port to listen on for HTTP and Websockets
Do Forever
   rc = WSService( !global.!context, 1000 ) -- timeout after 1 secs inactivity
End
-- shouldn't get here, but just in case...
Call WSDeleteContext !global.!context
Return 0

/*
 * Handle HTTP requests
 */
HTTPCallback: Procedure Expose (!globalv)
Parse Arg wsi, reason, file
lenhome = Length( !global.!server_home )
headers = 'Access-Control-Allow-Origin: *' || !global.!crlf
Select
   When reason = !REXXWS.!HTTP_REQUEST Then
      Do
         If file = '/' Then file = 'index.html'
         -- we have been given a file to serve in 'text'
         If Left( file, 1 ) = '/' Then tmpfile = !global.!server_home || Substr( file, 2 )
         Else tmpfile = !global.!server_home || file
         fullfile = Stream( tmpfile, 'C', 'QUERY EXISTS' )
         Parse Value Stream( fullfile, 'C', 'FSTAT' ) With . . . . . . . filetype
         Select
            When fullfile = '' Then
               Do
                  -- file doesn't exist
                  Call WSReturnHttpStatus wsi, !REXXWS.!HTTP_STATUS_NOT_FOUND
                  Call DoLog 'Request for' file 'not found', 1
                  Return 1
               End
            When Left( fullfile, lenhome ) \= !global.!server_home Then
               Do
                  -- file is not inside our home directory
                  Call WSReturnHttpStatus wsi, !REXXWS.!HTTP_STATUS_FORBIDDEN
                  Call DoLog 'Request for' file 'forbidden', 1
                  Return 1
               End
            When filetype = 'Directory' Then
               Do
                  Call WSReturnHttpStatus wsi, !REXXWS.!HTTP_STATUS_FORBIDDEN
                  Call DoLog 'Request for directory' file 'forbidden', 1
                  Return 1
               End
            When filetype = 'RegularFile' Then
               Do
                  -- file exists and is inside our home directory
                  mimetype = WSGetMimeType( file )
                  rc = WSServeHttpFile( wsi, fullfile, mimetype, headers )
                  Call DoLog 'Sent' fullfile 'as' mimetype 'rc:' rc, 1
                  Return rc
               End
            Otherwise
               Do
                  Call WSReturnHttpStatus wsi, !REXXWS.!HTTP_STATUS_FORBIDDEN
                  Call DoLog 'Request for unknown file type' filetype 'for' file 'forbidden', 1
                  Return 1
               End
         End
      End
   Otherwise
      Do
         Call DoLog 'Unhandled reason:' WSGetReasonText( reason ) || '('reason')'
      End
End
Return 0

ControlCallback: Procedure Expose (!globalv)
Parse Arg wsi, reason, text
rcode = 0 -- ensure websocket stays alive
Select
   When reason = !REXXWS.!WS_CONNECTED Then
      Do
         Call AddWebsocket wsi, 'control'
      End
   When reason = !REXXWS.!WS_CLOSED Then
      Do
         Call RemoveWebsocket wsi, 'control'
      End
   When reason = !REXXWS.!WS_RECEIVE Then
      Do
         Call DoLog 'Message from Control client('wsi'):' text, 1
         _json = JSONParse(text)
         id = ValidateJSON( wsi, _json )
         If id \= '' Then
            Do
               Select
                  When id = 00 Then Call SetDesktop wsi, _json
                  When id = 01 Then Call SetSlide wsi, _json
                  When id = 02 Then Call SetTab wsi, _json
                  When id = 20 Then Call GetCursor wsi, _json
                  When id = 21 Then Call SetCursor wsi, _json
                  When id = 22 Then Call ClickMouseButton wsi, _json
                  When id = 80 Then Call Hit wsi, _json
                  Otherwise Nop
               End
               Call JSONDelete _json
            End
      End
   When reason = !REXXWS.!SERVER_WRITEABLE Then
      Do
         rcode = WriteWebSocket( wsi )
      End
   Otherwise
      Do
         Call DoLog 'Unhandled reason:' WSGetReasonText( reason ) || '('reason')'
      End
End
Return rcode

SetSlide: Procedure Expose (!globalv)
Parse Arg wsi, _json
slide = JSONGetObjectItemValue( _json, 'slide' )
If slide = 0 Then !global.!slide = 0
Else !global.!slide += slide
Call SendSlideChange wsi, !global.!slide
Return

SetDesktop: Procedure Expose (!globalv)
Parse Arg wsi, _json
desktop = JSONGetObjectItemValue( _json, 'desktop' )
Address System 'xdotool set_desktop' desktop
-- get the current mouse location and send
Return

SetTab: Procedure Expose (!globalv)
Parse Arg wsi, _json
tab = JSONGetObjectItemValue( _json, 'tab' )
If tab = 1 Then key = 'ctrl+Page_Down'
Else key = 'ctrl+Page_Up'
Address System 'xdotool key' key
Return

SetCursor: Procedure Expose (!globalv)
Parse Arg wsi, _json
xperc = Min( JSONGetObjectItemValue( _json, 'xperc' ), 100 )
yperc = Min( JSONGetObjectItemValue( _json, 'yperc' ), 100 )
drag = JSONGetObjectItemValue( _json, 'drag' )
drop = JSONGetObjectItemValue( _json, 'drop' )
xpos = !global.!screenwidth * xperc / 100
ypos = !global.!screenheight * yperc / 100
If drag & \drop Then Address System 'xdotool mousedown 1'
If xperc >0 & yperc > 0 Then Address System 'xdotool mousemove' xpos ypos
If drop & drag Then Address System 'xdotool mouseup 1'
Return

ClickMouseButton: Procedure Expose (!globalv)
Parse Arg wsi, _json
button = JSONGetObjectItemValue( _json, 'button' )
Address System 'xdotool click' button
Return

SendSlideChange:
Parse Arg wsi, slide
page = 'slide-'Right( slide, 2, '0' )'.html'
If Stream( page, 'C', 'QUERY EXISTS' ) = '' Then
   Do
      Call QueueResponseJSON wsi, '99', CreateErrorObject( '99', 'Requested slide does not exist' )
      Return
   End
Call DoLog 'SendSlideChange:' page, 1
_root = JSONCreateObject( 'page', page)
Call QueueResponseJSON !global.!slidewsi, '01', _root
Return

Hit: Procedure Expose (!globalv)
Parse Arg wsi, _json
type = JSONGetObjectItemValue( _json, 'type' )
If type \= 'leave' Then
   Do
      _json = TranslateKeyCode( _json )
   End
lastq = RxQueue('G')
Call RxQueue 'S', 'the@localhost'
resp = JSONStringify( _json )
Say '---->' resp
Queue resp
Call RxQueue 'S', lastq
Return

/*
 * The callback for the slide presentation client
 * The only reason we have this is so we can get the web socket
 * to tell the client which slide to load
 */
SlideCallback: Procedure Expose (!globalv)
Parse Arg wsi, reason, text
rcode = 0 -- ensure websocket stays alive
Select
   When reason = !REXXWS.!WS_CONNECTED Then
      Do
         !global.!slidewsi = wsi
         Call AddWebsocket wsi, 'slide'
      End
   When reason = !REXXWS.!WS_CLOSED Then
      Do
         !global.!slidewsi = 0
         Call RemoveWebsocket wsi, 'slide'
      End
   When reason = !REXXWS.!WS_RECEIVE Then
      Do
         -- this should never be reached
         Call DoLog 'Message from Slide client('wsi'):' text, 1
      End
   When reason = !REXXWS.!SERVER_WRITEABLE Then
      Do
         rcode = WriteWebSocket( wsi )
      End
   Otherwise
      Do
         Call DoLog 'Unhandled reason:' WSGetReasonText( reason ) || '('reason')'
      End
End
Return rcode

/*
 * Common procedure to add a client connection
 */
AddWebsocket: Procedure Expose (!globalv)
Parse Arg wsi, service
Call DoLog service'('wsi') connection:', 1
-- create a new stack for each client
Call RxQueue 'C', 'Q'wsi
Return

/*
 * Common procedure to remove a client connection
 */
RemoveWebsocket: Procedure Expose (!globalv)
Parse Arg wsi, service
-- delete our stack for the client
Call RxQueue 'D', 'Q'wsi
Return

/*
 * Common procedure to write a buffer full of data on a client's stack to the client
 */
WriteWebSocket: Procedure Expose (!globalv)
Parse Arg wsi
Call DoLog 'WSI now writeable:' wsi
-- we can now write something; let's just write one line from stack
Call RxQueue 'S', 'Q'wsi
num = Queued()
If num \= 0 Then
   Do
      Parse Pull line
      rc = WSWrite( wsi, line, !REXXWS.!WRITE_TEXT )
      num = num - 1
   End
-- If we have more to write, enable this callback
If num \= 0 Then Call WSSetCallbackOnWriteable wsi
Return 0

/*
 * Ensure incoming JSON is valid and consists of an object with an "id" property
 */
ValidateJSON: Procedure Expose (!globalv)
Parse Arg wsi, json
If json = '' & !REXXJSON.!JSONCODE \= 0 Then
   Do
      Call DoLog 'Message from client('wsi') is not valid JSON;' !REXXJSON.!JSONERRM, 1
      Return ''
   End
If JSONDataType( json ) \= 'object' Then
   Do
      Call DoLog 'Message from client('wsi') is not recognised JSON; ignoring', 1
      Call JSONDelete json
      Return ''
   End
id = JSONGetObjectItemValue( json, 'id' )
If id = '' & !REXXJSON.!JSONCODE \= 0 Then
   Do
      Call DoLog 'Message from client('wsi') is not recognised JSON (no id property); ignoring', 1
      Call JSONDelete json
      Return ''
   End
Return id

-- puts the response on the client queue, logs it and triggers the writeable callback
QueueResponseJSON: Procedure Expose (!globalv)
Parse Arg wsi, id, _root -- _root is an object
If JSONDatatype( _root ) \= 'object' Then
   Do
      Say 'data is not a JSON object:' _root
      Exit 1
   End
Call JSONAddStringToObject _root, 'id', id
resp = JSONStringify( _root )
Call JSONDelete _root
Call RxQueue 'S', 'Q'wsi
Queue resp
Call DoLog 'Returning to' wsi':' resp, 1
Call WSSetCallbackOnWriteable wsi
Return

CreateErrorObject: Procedure Expose (!globalv)
Parse Arg id, error
Return JSONCreateObject( 'id', id, 'error', error )

DoLog: Procedure Expose (!globalv)
Parse Arg str, force
If force = 1 | !global.!debuglog = 1 Then
   Do
      Say str
   End
Return

-- type = 'key'
TranslateKeyCode: Procedure Expose (!globalv)
Parse Arg _json
keycode = JSONGetObjectItemValue( _json, 'text' )
ctrl = JSONGetObjectItemValue( _json, 'ctrl' )
alt = JSONGetObjectItemValue( _json, 'alt' )
shift = JSONGetObjectItemValue( _json, 'shift' )
Select
   When ctrl = 1 Then
      Do
         prefix = 'C-'
      End
   When alt = 1 Then
      Do
         prefix = 'A-'
      End
   When shift = 1 Then
      Do
         prefix = 'S-'
      End
   Otherwise prefix = ''
End
key = !key.keycode
If Words( key ) = 2 Then
   Do
      If shift = 1 Then key = Word( key, 1 )
      Else key = Word( key, 2 )
      prefix = ''
   End
Else
   Do
   End
keycode = prefix||key
Call JSONDeleteItemFromObject _json, 'text'
Call JSONAddStringToObject _json, 'text', keycode
Return _json

/*
 * Handle shutting down server
 */
HaltProgram:
Call WSDeleteContext !global.!context
Exit 0

Novalue:
Call Abort 'Novalue from line:' sigl 'Variable:' Condition('D'), 1
Return

Abort:
Signal Off Syntax
Signal Off Novalue
Parse Arg msg
Say msg
Exit 1

/*
 * Create a mapping of button numbers in keyboard.html to
 * THE keynames. US Keyboard layout
 */
SetupKeyCodes: Procedure Expose (!globalv)
!key. = ''
!key.0 = 105
!key.2 = 'ESC'
!key.4 = 'F1'
!key.5 = 'F2'
!key.6 = 'F3'
!key.7 = 'F4'
!key.9 = 'F5'
!key.10 = 'F6'
!key.11 = 'F7'
!key.12 = 'F8'
!key.14 = 'F9'
!key.15 = 'F10'
!key.16 = 'F11'
!key.17 = 'F12'
!key.23 = '~ `'
!key.24 = '! 1'
!key.25 = '@ 2'
!key.26 = '# 3'
!key.27 = '$ 4'
!key.28 = '% 5'
!key.29 = '^ 6'
!key.30 = '& 7'
!key.31 = '* 8'
!key.32 = '( 9'
!key.33 = ') 0'
!key.34 = '_ -'
!key.35 = '+ ='
!key.36 = 'BKSP'
!key.38 = 'INS'
!key.39 = 'HOME'
!key.40 = 'PGUP'
!key.43 = 'TAB'
!key.44 = 'Q'
!key.45 = 'W'
!key.46 = 'E'
!key.47 = 'R'
!key.48 = 'T'
!key.49 = 'Y'
!key.50 = 'U'
!key.51 = 'I'
!key.52 = 'O'
!key.53 = 'P'
!key.54 = '{ ['
!key.55 = '} ]'
!key.56 = '| \'
!key.58 = 'DEL'
!key.59 = 'END'
!key.60 = 'PGDN'
!key.64 = 'A'
!key.65 = 'S'
!key.66 = 'D'
!key.67 = 'F'
!key.68 = 'G'
!key.69 = 'H'
!key.70 = 'J'
!key.71 = 'K'
!key.72 = 'L'
!key.73 = ': ;'
!key.74 = '"' "'"
!key.75 = 'ENTER'
!key.79 = 'Z'
!key.80 = 'X'
!key.81 = 'C'
!key.82 = 'V'
!key.83 = 'B'
!key.84 = 'N'
!key.85 = 'M'
!key.86 = '< ,'
!key.87 = '> .'
!key.88 = '? /'
!key.92 = 'CURU'
!key.98 = 'SPACE'
!key.103 = 'CURL'
!key.104 = 'CURD'
!key.105 = 'CURR'
Return

GetScreenDimensions: Procedure Expose (!globalv)
Address System 'xrandr | head -1' With Output Stem screen.
Parse Var screen.1 . 'current' width 'x' height ',' .
!global.!screenwidth = Strip( width )
!global.!screenheight = Strip( height )
Return
