#!/usr/bin/regina
/*
 * This program runs as a basic HTML server
 */
Trace O
Signal on Halt Name HaltProgram
Signal On Novalue
Numeric Digits 30
!globalv = '!REXXWS. !global.'
-- constants
!global.!debuglog = 0
Call RxFuncAdd 'WSLoadFuncs', 'rexxws', 'WSLoadFuncs'
Call WSLoadFuncs
Parse Arg _port _home
If Datatype( _port ) \= 'NUM' Then Call Abort 'Supplied port number is not numeric:' _port
server_home = Stream( _home, 'C', 'QUERY EXISTS' )
If server_home = '' Then Call Abort 'Supplied server home does not exist:' _home
!global.!port = _port
!global.!server_home = server_home'/'

Parse Source os . me
Call Directory !global.!server_home
If os = 'WIN32' | os = 'WIN64' Then !global.!sep = '\'
Else !global.!sep = '/'
!global.!crlf = '0d0a'x

Call DoLog 'Starting ws-server.rexx: Rexx/WS version:' WSVariable( 'VERSION' ) 'libwebsockets version:' WSVariable( 'WSVERSION' ), 1
Call DoLog 'Listening on port' !global.!port 'in' !global.!server_home, 1
If !global.!debuglog = 1 Then logging = WSOr(!REXXWS.!LOG_DEBUG,!REXXWS.!LOG_INFO,!REXXWS.!LOG_NOTICE,!REXXWS.!LOG_ERR,!REXXWS.!LOG_WARN )
Else logging = 0
Call WSSetLogLevel( logging )

rc = WSAddProtocol( !REXXWS.!PROTOCOL_HTTP, 'HTTPCallback' )

options = 0
!global.!context = WSCreateContext( !global.!port, '', '', options ) -- port to listen on for HTTP and Websockets
Do Forever
   rc = WSService( !global.!context, 1000 ) -- timeout after 1 secs inactivity
   /* here we can do housekeeping while no requests to service */
End
-- shouldn't get here, but just in case...
Call WSDeleteContext !global.!context
Return 0

/*
 * Handle HTTP requests
 */
HTTPCallback: Procedure Expose (!globalv)
Parse Arg wsi, reason, file
lenhome = Length( !global.!server_home )
headers = 'Access-Control-Allow-Origin: *' || !global.!crlf
Select
   When reason = !REXXWS.!HTTP_REQUEST Then
      Do
         If file = '/' Then file = 'index.html'
         -- we have been given a file to serve in 'text'
         If Left( file, 1 ) = '/' Then tmpfile = !global.!server_home || Substr( file, 2 )
         Else tmpfile = !global.!server_home || file
         fullfile = Stream( tmpfile, 'C', 'QUERY EXISTS' )
         Parse Value Stream( fullfile, 'C', 'FSTAT' ) With . . . . . . . filetype
         Select
            When fullfile = '' Then
               Do
                  -- file doesn't exist
                  Call WSReturnHttpStatus wsi, !REXXWS.!HTTP_STATUS_NOT_FOUND
                  Call DoLog 'Request for' file 'not found', 1
                  Return 1
               End
            When Left( fullfile, lenhome ) \= !global.!server_home Then
               Do
                  -- file is not inside our home directory
                  Call WSReturnHttpStatus wsi, !REXXWS.!HTTP_STATUS_FORBIDDEN
                  Call DoLog 'Request for' file 'forbidden', 1
                  Return 1
               End
            When filetype = 'Directory' Then
               Do
                  Call WSReturnHttpStatus wsi, !REXXWS.!HTTP_STATUS_FORBIDDEN
                  Call DoLog 'Request for directory' file 'forbidden', 1
                  Return 1
               End
            When filetype = 'RegularFile' Then
               Do
                  -- file exists and is inside our home directory
                  mimetype = WSGetMimeType( file )
                  rc = WSServeHttpFile( wsi, fullfile, mimetype, headers )
                  Call DoLog 'Sent' fullfile 'as' mimetype 'rc:' rc, 1
                  Return rc
               End
            Otherwise
               Do
                  Call WSReturnHttpStatus wsi, !REXXWS.!HTTP_STATUS_FORBIDDEN
                  Call DoLog 'Request for unknown file type' filetype 'for' file 'forbidden', 1
                  Return 1
               End
         End
      End
   Otherwise
      Do
         Call DoLog 'Unhandled reason:' WSGetReasonText( reason ) || '('reason')'
      End
End
Return 0

DoLog: Procedure Expose (!globalv)
Parse Arg str, force
If force = 1 | !global.!debuglog = 1 Then
   Do
      Say str
   End
Return

/*
 * Handle shutting down server
 */
HaltProgram:
Call WSDeleteContext !global.!context
Exit 0

Novalue:
Call Abort 'Novalue from line:' sigl 'Variable:' Condition('D'), 1
Return

Abort:
Signal Off Syntax
Signal Off Novalue
Parse Arg msg
Say msg
Exit 1
